(defpackage test-prop-logic
  (:use 
   cl
   utils
   prod-set
   set
   prop-logic))

(in-package test-prop-logic)


(setf f1 'foo)
(setf f2 (disjoin 'foo 'bar))
(setf f3 (conjoin 'foo (negate '(bar 4))))
(setf f4 (disjoin t (conjoin 'foo 'bar) (conjoin (negate 'bar) (negate 'baz))))
(setf f5 (negate 'qux))
(setf f6 (conjoin 'foo (disjoin 'bar 'baz)))


(do-boolean-tests 
    "Propositional formula types"
  (is-formula t) t
  (is-formula nil) t
  (is-formula 'foo) t
  (is-formula 3) nil
  (is-formula '(a b 6)) t 
  (is-formula '(a 4 (b 4))) nil
  (is-formula '(or)) t
  (is-formula '(and nil (or (and foo t) (and (not (bar 3)))) (not (and bar (or baz nil))))) t
  (is-formula '(and nil (or (and foo t) (and (not (bar 3)))) (not nil (and bar (or baz nil))))) nil
  (is-formula '(and nil (or (and (foo) t) (and (not bar))) (not nil (and bar (or baz nil))))) nil
  (is-literal f1) t
  (is-literal t) nil
  (is-literal nil) nil
  (is-literal f5) t
  (is-literal f2) nil
  (is-literal f3) nil
  (is-dnf-clause f1) t
  (is-dnf-clause f2) nil
  (is-dnf-clause f3) t
  (is-dnf-clause t) t
  (is-dnf-clause nil) t
  (is-dnf-clause f4) nil
  (is-dnf-formula t) t
  (is-dnf-formula nil) t
  (is-dnf-formula f1) t
  (is-dnf-formula f2) t
  (is-dnf-formula f3) t
  (is-dnf-formula f4) t
  (is-dnf-formula f5) t
  (is-dnf-formula f6) nil)
  

(setf f1 '(and (foo x (a 3)) (not (bar y)) (not (or (x (b 2) y))))
      b '((x . y) (y . 24)))


(do-tests
    "Substitution"
  (setf f2 (bind f1 b)) '(and (foo y (a 3)) (not (bar 24)) (not (or (x (b 2) 24))))
  f1 '(and (foo x (a 3)) (not (bar y)) (not (or (x (b 2) y))))
  (bind 'x b) 'x
  (bind '(and t nil (foo y)) b) '(and t nil (foo 24)))

(let ((s1 nil)
      (s2 '((foo) (bar) (baz)))
      (s3 '((baz) (qux a 3 4) (foo))))
  (do-boolean-tests
      "Proposition formula holds function"
    (holds s1 t) t
    (holds s3 t) t
    (holds s1 nil) nil
    (holds s2 nil) nil
    (holds s1 'bar) nil
    (holds s2 'bar) t
    (holds s3 'bar) nil
    (holds s1 '(foo)) nil
    (holds s3 '(foo)) t
    (holds s1 '(not qux)) t
    (holds s2 '(not qux)) t
    (holds s3 '(not (qux a 3 4))) nil
    (holds s2 '(or (and foo bar) (not qux))) t
    (holds s3 '(or (and foo bar) (not (qux a 3 4)))) nil
    (holds s1 '(or (and foo bar) (not qux))) t))





(let ((a (make-prop-set 'foo 3 '(x y))))
  (do-boolean-tests
      "Proposition sets"
    (member? '(foo 2 x) a) t
    (member? '(foo 3 x) a) nil
    (member? '(foo 1) a) nil
    (member? '(bar 2 y) a) nil))




(setf f1 '(or (and foo bar) nil (not baz) (and (not bar) foo)))
(setf f2 '(or (and baz qux) (and (not qux) (not foo))))



(do-tests
    "DNF operations"
  (simplify-dnf t) t
  (simplify-dnf nil) nil
  (simplify-dnf 'foo) 'foo
  (simplify-dnf '(not foo)) '(not foo)
  (simplify-dnf '(and foo bar foo)) '(and (bar) (foo))
  (simplify-dnf '(and foo (bar) foo (not bar))) nil
  (simplify-dnf '(or foo bar (and foo qux (not qux)))) '(or foo bar)
  (simplify-dnf '(or foo bar (and foo qux (not qux)) t)) t
  (sort-dnf (simplify-dnf (dnf-and f1 f2)))
  '(or (and (bar) (baz) (foo) (qux)) (and (not (bar)) (baz) (foo) (qux)) (and (not (baz)) (not (foo)) (not (qux))))
  (sort-dnf (dnf-not '(and foo (not bar)))) '(or bar (not foo))
  (sort-dnf (dnf-not '(or (and foo bar) (not baz)))) '(or (and (not bar) baz) (and baz (not foo)))
  
)



(do-boolean-tests
    "DNF implication"
  (implies '(or (and (foo 4 5) bar) (and bar (not baz))) '(foo 4 5))
  nil
  
  (implies '(or (and foo bar) (and baz (not foo))) 'foo)
  nil
  
  (implies '(or (and foo bar) (and baz (not foo))) t)
  t
  
  (implies '(or (and foo bar) (and baz foo)) nil)
  nil
  
  (implies '(or (and foo bar) (and baz (not foo))) '(or bar baz))
  t
  
  (implies '(or (and foo bar) (and baz (not foo))) '(or foo bar))
  nil
  
  (implies '(or (and (foo x) bar) (and baz (foo x))) '(foo x))
  t)

      

(setf s (make-dnf-set f1 '((foo) (bar) (qux) (baz)))
      s2 (make-instance '<dnf-set> :formula f2 :props '((foo) (bar) (qux) (baz)) :simplified t)
      s3 (intersect s s2)
      s4 (binary-union s s2)
      f3 '(or (and (not qux) (not foo)) (and baz qux))
      s5 (make-instance '<dnf-set> :formula f3 :props '((foo) (bar) (qux) (baz)) :simplified t))



(do-tests
    "DNF sets"
  (member? '((foo) (bar) (qux)) s) t
  (member? '((baz)) s) nil
  (member? '((foo) (bar) (qux)) s3) nil
  (member? '((foo) (baz) (qux)) s3) t
  (member? '((foo) (bar) (qux)) s4) t
  (member? '((foo) (baz) (qux)) s4) t
  (size-must-exceed s2 4) nil
  (size-must-exceed s2 3) t
  (set-eq s2 s5) t
  (set-eq s2 s) nil
  )
