(defpackage lookahead-submission
  (:use lookahead
	utils
	tree
	blocks
	cl
	mapping
	env-user
	set
	prop-logic)
  (:import-from lookahead
		precond
		succ-state
		high-level-actions
		action-type
		is-primitive-sequence
		instantiate
		complete-desc
		sound-desc
		complete-result
		sound-result
		succeeds-complete
		get-immediate-refinements
		succeeds-sound
		regress-clause-nstrips
		top-level-plans
		))

(in-package lookahead-submission)

;; See also lookahead/test-lookahead.lisp domain d for the running example

(setf 
    d2 (make-blocks-world-with-ceiling 4 3 '((a 0 1) (baz 2 1) (c 2 2))
				       '(2 3) '(and (on c a)))
    u2 (make-unguided-blocks-hierarchy d2)
    q2 (make-instance '<prop-abstract-planning-problem>
	 :planning-problem d2 :hierarchy u2
	 :complete-desc-schemas (make-complete-descriptions d2)
	 :sound-desc-schemas (make-sound-descriptions d2))
    q2c (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d2 :hierarchy u2
	  :complete-desc-schemas (make-complete-descriptions d2)
	  :sound-desc-schemas (make-dummy-sound-descriptions d2))
    q2s (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d2 :hierarchy u2
	  :hset-desc-schemas (make-complete-descriptions d2)
	  :complete-desc-schemas (make-dummy-complete-descriptions d2)
	  :sound-desc-schemas (make-sound-descriptions d2))
    q2h (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d2 :hierarchy u2
	  :hset-desc-schemas (make-complete-descriptions d2)
	  :complete-desc-schemas (make-dummy-complete-descriptions d2)
	  :sound-desc-schemas (make-dummy-sound-descriptions d2))
    h2 (make-blocks-hierarchy d2)
    flat2 (make-flat-hierarchy d2)
    p2 (make-instance '<prop-abstract-planning-problem>
	 :planning-problem d2 :hierarchy h2
	 :complete-desc-schemas (make-complete-descriptions d2)
	 :sound-desc-schemas (make-sound-descriptions d2))
    p2s (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d2 :hierarchy h2
	  :hset-desc-schemas (make-complete-descriptions d2)
	  :complete-desc-schemas (make-dummy-complete-descriptions d2)
	  :sound-desc-schemas (make-sound-descriptions d2))
    p2f (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d2 :hierarchy flat2)
    p2c (make-instance '<prop-abstract-planning-problem>
	  :hierarchy h2 :planning-problem d2
	  :complete-desc-schemas (make-complete-descriptions d2)
	  :sound-desc-schemas (make-dummy-sound-descriptions d2))
    p2h (make-instance '<prop-abstract-planning-problem>
	  :hierarchy h2 :planning-problem d2
	  :hset-desc-schemas (make-complete-descriptions d2)
	  :complete-desc-schemas (make-dummy-complete-descriptions d2)
	  :sound-desc-schemas (make-dummy-sound-descriptions d2))
    d3 (make-blocks-world-with-ceiling 4 4 '((a 0 1) (baz 2 1) (c 2 2))
				       '(2 3) '(and (on c a) (on baz c)))
    flat3 (make-flat-hierarchy d3)
    u3 (make-unguided-blocks-hierarchy d3)
    q3 (make-instance '<prop-abstract-planning-problem>
	 :planning-problem d3 :hierarchy u3
	 :complete-desc-schemas (make-complete-descriptions d3)
	 :sound-desc-schemas (make-sound-descriptions d3))
    q3s (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d3 :hierarchy u3
	  :complete-desc-schemas (make-dummy-complete-descriptions d3)
	  :sound-desc-schemas (make-sound-descriptions d3))
    q3h (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d3 :hierarchy u3
	  :complete-desc-schemas (make-dummy-complete-descriptions d3)
	  :sound-desc-schemas (make-dummy-sound-descriptions d3))
    q3c (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d3 :hierarchy u3
	  :complete-desc-schemas (make-complete-descriptions d3)
	  :sound-desc-schemas (make-dummy-sound-descriptions d3))
    d4 (make-blocks-world-with-ceiling 6 4 '((a 0 1) (baz 2 1) (c 2 2))
				       '(2 4) '(and (on baz :t3) (on c baz)))
    flat4 (make-flat-hierarchy d4)
    u4 (make-unguided-blocks-hierarchy d4)
    q4 (make-instance '<prop-abstract-planning-problem>
	 :planning-problem d4 :hierarchy u4
	 :complete-desc-schemas (make-complete-descriptions d4)
	 :sound-desc-schemas (make-sound-descriptions d4))
    q4c (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d4 :hierarchy u4
	  :complete-desc-schemas (make-complete-descriptions d4)
	  :sound-desc-schemas (make-dummy-sound-descriptions d4))
    q4h (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d4 :hierarchy u4
	  :complete-desc-schemas (make-dummy-complete-descriptions d4)
	  :sound-desc-schemas (make-dummy-sound-descriptions d4))
    p4f (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d4 :hierarchy flat4)
    d5 (make-blocks-world-with-ceiling 8 5 '((a 0 1) (baz 2 1) (c 2 2) (qux 4 1) )
				       '(3 3) '(and (on baz :t3) (on c baz) (on a :t1)))
    u5 (make-unguided-blocks-hierarchy d5)
    q5 (make-instance '<prop-abstract-planning-problem>
	 :planning-problem d5 :hierarchy u5
	 :complete-desc-schemas (make-complete-descriptions d5)
	 :sound-desc-schemas (make-sound-descriptions d5))
    q5c (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d5 :hierarchy u5
	  :complete-desc-schemas (make-complete-descriptions d5)
	  :sound-desc-schemas (make-dummy-sound-descriptions d5))
    d6 (make-blocks-world-with-ceiling 4 4 '((a 0 1) (baz 2 1) (c 2 2))
				       '(2 3) '(and (on c :t3) (on baz c) (on a baz)))
    u6 (make-unguided-blocks-hierarchy d6)
    q6 (make-instance '<prop-abstract-planning-problem>
	 :planning-problem d6 :hierarchy u6
	 :complete-desc-schemas (make-complete-descriptions d6)
	 :sound-desc-schemas (make-sound-descriptions d6))
    q6c (make-instance '<prop-abstract-planning-problem>
	  :planning-problem d6 :hierarchy u6
	  :complete-desc-schemas (make-complete-descriptions d6)
	  :sound-desc-schemas (make-dummy-sound-descriptions d6))
    )




(defvar *init-depth-limit* 2)

(defun run-first (p d &optional (n 1))
  (time (dotimes (i n) (hierarchical-forward-search p :first-action-only t :init-depth-limit *init-depth-limit*))))

(defun run (p d &optional (n 1))
  (time (dotimes (i n) (assert (succeeds? d (hierarchical-forward-search p :init-depth-limit *init-depth-limit*))))))


;; Commented out runs are the ones that take too long

(defvar *num-trials* 5)

(setf *refinement-set-type* ':queue)
(format t "~&Algorithm A")
(run-first q2 d2 *num-trials*)
(run-first q4 d4 *num-trials*)
(run-first q5 d5 *num-trials*)

(format t "~&Algorithm SC")
(run q2 d2 *num-trials*)
(run q4 d4 *num-trials*)
(run q5 d5 *num-trials*)

(setf *refinement-set-type* ':stack)

(format t "~&Algorithm C")
(run q2c d2 *num-trials*)
(run q4c d4 *num-trials*)
;; (run q5c d5 *num-trials*) 

(format t "~&Algorithm H")
(run q2h d2 *num-trials*)
;; (run q4h d4 *num-trials*)

(format t "~&Algorithm F")
(run p2f d2 *num-trials*)
;; (run p4f d4 *num-trials*)
