(defpackage exec
  (:documentation "
Contains code relating to creating standalone executables for parts of the ALisp code.
")
  (:use
   asdf
   cl)
  (:export
   create-alisp-executable)
  )
  

(in-package exec)

(defclass list-op (operation) 
  ((vec :initarg :vec :accessor vec)))

(defmethod perform ((o list-op) (c cl-source-file))
  (map nil
    #'(lambda (f)
	(vector-push-extend (compile-file-pathname f) (vec o)))
    (input-files o c)))

(defmethod component-depends-on ((o list-op) (c component))
  (let ((what-would-load-op-do (cdr (assoc 'load-op
                                           (slot-value c 'asdf::in-order-to)))))
    (mapcar #'(lambda (dep)
		(if (eq (car dep) 'load-op)
		    (cons 'list-op (cdr dep))
                  dep))
            what-would-load-op-do)))

(defmethod create-alisp-executable (target script)
  "create-alisp-executable TARGET SCRIPT
Creates an executable TARGET that runs SCRIPT, then terminates."
  (unwind-protect 
      (let ((files (let ((v (make-array 0 :adjustable t :fill-pointer 0)))
		     (asdf:operate 'list-op 'hrl :vec v :force t)
		     (coerce v 'list)))
	    (dir (make-pathname :directory `(:relative ,target))))
	(excl:delete-directory-and-files dir :if-does-not-exist :ignore)
	(sleep 1)
	(excl:generate-executable target (append files (list (pathname script)))
				  :include-compiler nil))
    ))