(in-package utils)

(defconstant *default-tail-length* 10)

(defun stail (s &optional (tail-length *default-tail-length*))
  "stail SEQ &optional (LENGTH 10).  Return last LENGTH elements of SEQ."
  (if (> (length s) tail-length)
      (subseq s (- (length s) tail-length))
    s))

(defun slast (s)
  "slast SEQ.  Last element of SEQ."
  (assert (> (length s) 0))
  (elt s (1- (length s))))


(defmacro def-seq-accessor (name pos)
  (let ((s (gensym)))
    `(defun ,name (,s) (elt ,s ,pos))))

(def-seq-accessor sfirst 0)
(def-seq-accessor ssecond 1)
(def-seq-accessor sthird 2)
(def-seq-accessor sfourth 3)


(defgeneric is-sorted (l &optional pred)
  (:documentation "is-sorted SEQUENCE &OPTIONAL (PRED #'<=).  Do every pair of adjacent elements in SEQUENCE satisfy PRED?  (We're assuming PRED is transitive)")
  (:method ((l list) &optional (pred #'<=))
	   (loop
	       for rem on l
	       always (or (not (cdr rem)) (funcall pred (car rem) (cadr rem)))))
  (:method ((l vector) &optional (pred #'<=))
	   (loop
	       for i below (1- (length l))
	       always (funcall pred (aref l i) (aref l (1+ i))))))


(defun extract-subsequence (seq init skip &key (result-type 'vector) &aux (n (length seq)))
  "extract-subsequence SEQUENCE INITIAL SKIP
Return the sequence consisting of SEQUENCE(INITIAL), SEQUENTIAL(INITIAL+SKIP), ..."

  (ecase result-type
    (list (loop
	      for x from init below n by skip
	      collect (elt seq x)))
    (vector (loop
		with a = (make-array (ceiling n skip) :fill-pointer 0)
		for x from init below n by skip
		do (vector-push (elt seq x) a)
		finally (return a)))))


(defun round-sequence (l &optional (d 2))
  "round-sequence S &optional (D 2)"
  (map (if (listp l) 'list 'vector)
    #'(lambda (x) (round-decimal x d)) l))



(defun changepoints (seq &key (test #'eql) (key #'identity) (start 0) (end (length seq)))
  "changepoints SEQUENCE &key (TEST #'eql) (KEY #'identity)

Suppose you have the sequence (0 0 1 2 1 1 2)
This function would return the list ((0 . 0) (1 . 2) (2 . 3) (1 . 4) (2 . 6))
The meaning of a pair (i . j) in this list is that, starting from position j until the j value of the next pair, the elements have key values equal to i."
  
  (etypecase seq
    (list 
     (loop
	 with prev-k = nil
		    
	 for i from start below end
	 for first-time = t then nil
	 for x in (nthcdr start seq)
	 for k = (funcall key x)

	 when (or first-time (not (funcall test k prev-k)))
	 collect (cons k i)
		 
	 do (setf prev-k k)))
     
    
    (vector
     (loop
	 with prev-k = nil
	 for i from start below end
	 for first-time = t then nil
	 for x = (aref seq i)
	 for k = (funcall key x)
	 when (or first-time (not (funcall test k prev-k)))
	 collect (cons k i)
		 
	 do (setf prev-k k)))))
      
			      








