(defpackage stat-gatherer
  (:documentation "Package for the <stat-gatherer> subclass of <rl-observer>.

Exports
-------
<stat-gatherer>
total-reward
num-episodes-started
num-episodes-finished
num-env-steps
reset-counts")
  (:use common-lisp
	rl-obs
	utils)
  (:export
   <stat-gatherer>
   total-reward
   num-episodes-started
   reset-counts
   num-episodes-finished
   num-env-steps))

(in-package stat-gatherer)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; class def
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass <stat-gatherer> (<rl-observer>)
  ((total-reward 
    :accessor rew 
    :reader total-reward 
    :initform 0)
   (num-episodes-started 
    :accessor num-ep-started 
    :reader num-episodes-started 
    :initform 0)
   (num-episodes-finished 
    :accessor num-ep-finished 
    :reader num-episodes-finished
    :initform 0)
   (num-env-steps 
    :accessor sg-num-env-steps 
    :reader num-env-steps 
    :initform 0)
   )
  
  (:documentation "Class for gathering stats during the execution of the main loop of reinforcement-learning, alisp, or concurrent alisp.   Right now counts total reward, num episodes started, num episodes ended, num env steps, num choice states.  Can add more stats in future, but make sure everything is O(1) space (so we can include this class as a default observer and not worry about memory).  The counts start at 0.  To reset them back to 0, use reset-counts.


"))


(defmethod print-object ((obs <stat-gatherer>) str)
  
  (format str "~&RL stats~&~a episodes started and ~a finished~&~a env steps and total reward is ~a"
	  (num-episodes-started obs) (num-episodes-finished obs) (num-env-steps obs) 
	  (total-reward obs)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; methods overridden from <rl-observer>
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun reset-counts (obs)
  
  (setf (num-ep-started obs) 0
	(num-ep-finished obs) 0
	(sg-num-env-steps obs) 0
	(rew obs) 0))

(defmethod inform-start-episode ((obs <stat-gatherer>) s)
  (declare (ignore s))
  (incf (num-ep-started obs)))

(defmethod inform-env-step ((obs <stat-gatherer>) act rew to term)
  (declare (ignore to act))
  (incf (rew obs) rew)
  (incf (sg-num-env-steps obs))
  (when term (incf (num-ep-finished obs))))

